
# This tutorial is about R parallelization and has been inspired from http://glennklockwood.com/di/R-para.php 

rm(list=ls()) #clear workspace

source("genData.R") #loading the data generation function
#install.packages("e1071", lib="/opt/conda/lib/R/library", repo="http://cran.us.r-project.org")
#install.packages("snow", lib="/opt/conda/lib/R/library", repo="http://cran.us.r-project.org")
#install.packages("mclust", lib="/opt/conda/lib/R/library", repo="http://cran.us.r-project.org")
#install.packages("doMC", lib="/opt/conda/lib/R/library", repo="http://cran.us.r-project.org")
#install.packages("foreach", lib="/opt/conda/lib/R/library", repo="http://cran.us.r-project.org")

library(e1071)
library(parallel)
library(foreach)
library(doMC)
library(mclust)
library(snow)
#genData

data=read.csv("data/iris_h.csv")
labels=data$Species
pairs(data[1:4], main = "Fisher 's Iris Data", pch = 21, bg = c("red", "green3", "blue")[unclass(iris$Species)])
head(data)
data$Species<-NULL
head(data)


# parameter matrix: each row is a combination of modelNames and G (see Mclust function)
modelNames=c("EII","EEI","EEE","VVV")
G=2:6
param=expand.grid(G=G,modelNames=modelNames) #12 combinations!
rm(G,modelNames)

# how many cores do we want to exploit during the parallelization?
ncores=ifelse(nrow(param)>16,yes=16,no=nrow(param))

parallel.function <- function(i) { Mclust( data=data, G=as.character(param$G[i]),
                                          modelNames=as.character(param$modelNames[i]),prior=priorControl()) }
# NB: the parallelization will not be about splitting the data, but about choosing the best parameters
# the data are hard-coded in the function, the variables here are "G"(1) and "modelNames"(2)
ncores

b=Sys.time()
results_list=mclapply((1:nrow(param)),parallel.function,mc.cores=ncores)
e=Sys.time()
mclapply_time=e-b
print(mclapply_time)
mclapply_result=results_list[[which.max(sapply(results_list,function(x){x$bic}))]]
rm(e,b,results_list)

ls(mclapply_result)
print(mclapply_result$G)
print(mclapply_result$modelName)

# matching the results with the actual labels
matchClasses(table(labels,mclapply_result$cl),method="exact")
# entropy of the result
mclapply_entropy=-sum((table(mclapply_result$cl)/length(mclapply_result$cl))*log10(table(mclapply_result$cl)/length(mclapply_result$cl)))
mclapply_entropy
# which obs have been incorrectly classified?
errors=labels
errors[classError(mclapply_result$cl,truth=labels)$misclassified]=2
errors[-classError(mclapply_result$cl,truth=labels)$misclassified]=1
plot(data[,1],data[,2],col=errors,main="misclassified observations (red dots)",xlab="x",ylab="y")

b=Sys.time()
Mclust(data,G=mclapply_result$G,modelNames=mclapply_result$modelName,prior=priorControl())
e=Sys.time()
print(e-b)
rm(b,e)


registerDoMC(ncores) #it is a necessary step in order to make foreach work in parallel

b=Sys.time()
results_list <- foreach(i = 1:nrow(param)) %dopar% parallel.function(i)
e=Sys.time()
doMC_result = results_list[[which.max(sapply(results_list,function(x){x$bic}))]]
doMC_time=e-b
print(doMC_time)
rm(e,b,results_list)

doMC_result
print(mclapply_result$G)
print(mclapply_result$modelName)
labels
doMC_result$cl
# matching the results with the actual labels
matchClasses(table(labels,doMC_result$cl),method="rowmax")
# entropy of the result
doMC_entropy=-sum((table(doMC_result$cl)/length(doMC_result$cl))*log10(table(doMC_result$cl)/length(doMC_result$cl)))
doMC_entropy


